/* Metageek WiSPY LibUSB interface 
 *
 * LibUSB userspace control for the Metageek WiSPY spectrum analyzer
 * http://www.metageek.net/
 *
 * Version 1.0 - Curses visualizer
 * Mike Kershaw/Dragorn <dragorn@kismetwireless.net>
 *
 * This code is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Extra thanks to Ryan Woodings @ Metageek for interface documentation
 */

#include <stdio.h>
#include <usb.h>
#include <curses.h>

#include "config.h"

#ifdef HAVE_LINUX_DEVDISCONNECT
/* Needed for our own tools */
#include <unistd.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <linux/usbdevice_fs.h>
#include <errno.h>
#endif

/* First-gen firmware */
#define METAGEEK_WISPY1_VID	0x04b4
#define METAGEEK_WISPY1_PID	0x0bad

/* Second-gen firmware */
#define METAGEEK_WISPY2_VID	0x1781
#define METAGEEK_WISPY2_PID	0x083e

#define TIMEOUT 3000

/* Hid get report from spec */
#define HID_GET_REPORT 0x01
#define HID_RT_FEATURE 0x03

/* maximum number of samples */
#define WISPY_SAMPLES 84
#define WISPY_MINSIG -65
#define WISPY_AVGNUM 3

/* Graph height ...  LINES - borders - 2 for chnum */
#define GRAPH_HEIGHT (LINES - 4)

/* Basic device initialize, grabs either version 1 or version 2 */
static struct usb_device *init_wispy_device() {
	struct usb_bus *bus;
	struct usb_device *dev;

	/* Libusb init */
	usb_init();
	usb_find_busses();
	usb_find_devices();

	/* Scan all the devices on all the busses */
	for (bus = usb_busses; bus; bus = bus->next) {
		for (dev = bus->devices; dev; dev = dev->next) {
			if (((dev->descriptor.idVendor == METAGEEK_WISPY1_VID) &&
				 (dev->descriptor.idProduct == METAGEEK_WISPY1_PID)) ||
				((dev->descriptor.idVendor == METAGEEK_WISPY2_VID) &&
				 (dev->descriptor.idProduct == METAGEEK_WISPY2_PID)))
				return dev;
		}
	}

	return NULL;
}

#ifdef HAVE_LINUX_DEVDISCONNECT
/* Libusb doesn't seem to always provide this, so we'll use our own, taken from the 
* usb_detatch_kernel_driver_np...
*
* THIS IS A HORRIBLE EVIL HACK THAT SHOULDN'T BE DONE, EVER
* 
*/
struct local_usb_ioctl {
	int ifno;
	int ioctl_code;
	void *data;
};

struct ghetto_libusb_devhandle {
	int fd;
	/* Nooo... so bad. */
};

int usb_detach_kernel_driver(struct usb_dev_handle *dev, int interface) {
	struct local_usb_ioctl command;
	struct ghetto_libusb_devhandle *gdev;

	command.ifno = interface;
	command.ioctl_code = USBDEVFS_DISCONNECT;
	command.data = NULL;

	gdev = (struct ghetto_libusb_devhandle *) dev;

	if (ioctl(gdev->fd, USBDEVFS_IOCTL, &command) < 0) {
		fprintf(stderr, "Could not detach kernel driver from interface %d: %s",
			interface, strerror(errno));
		return -1;
	}

	return 0;
}
#endif

int main(int argc, char *argv[]) {
	struct usb_device *dev;
	struct usb_dev_handle *wispy;
	int arg;
	char buf[8];
	int x, cursesdraw = 0, avgpos = 0;
	struct timeval tm;
	fd_set readset;

	/* previous samples */
	int lastsignal[WISPY_SAMPLES];
	int peaksignal[WISPY_SAMPLES];
	int lastpeaksignal[WISPY_SAMPLES];
	int avgsignal[WISPY_AVGNUM][WISPY_SAMPLES];
	char line[WISPY_SAMPLES + 1];

	/* curses windows */
	WINDOW *window;
	WINDOW *sigwin;
	WINDOW *chwin;

	memset(buf, 0, sizeof(char) * 8);
	for (x = 0; x < WISPY_SAMPLES; x++) {
		lastsignal[x] = 0;
		peaksignal[x] = WISPY_MINSIG;
		lastpeaksignal[x] = WISPY_MINSIG;
	}
	for (avgpos = 0; avgpos < WISPY_AVGNUM; avgpos++) {
		for (x = 0; x < WISPY_SAMPLES; x++)
			avgsignal[avgpos][x] = 0;
	}
	avgpos = 0;

	dev = init_wispy_device();
	if (!dev) {
		printf("Unable to find device.\n");
		return 1;
	}

	/* Open the wispy device */
	wispy = usb_open(dev);
	if (!wispy) {
		printf("Unable to open device\n");
		return 1;
	}

#if 1
#ifndef SYS_DARWIN
	/* Grab it like a keyboard interface, interface 0 */
	if (usb_claim_interface(wispy, 0) < 0) {
		printf("Unable to claim interface 0 -- are you root?\n"
			   "Is something else using it (another program or the kernel HID)?\n");
#ifdef HAVE_LINUX_DEVDISCONNECT
		printf("Using terrible hack to try to detach from the Linux HID...\n"
			   "This might not work at all\n");
		if (usb_detach_kernel_driver(wispy, 0) < 0) {
			printf("Failed to detatch from HID.\n");
			exit(1);
		}

		if (usb_claim_interface(wispy, 0) < 0) {
			printf("Seemed to detach from kernel but still couldn't claim interface.\n");
			exit(1);
		}

		printf("Detatched the USB interface from the kernel driver.\n");
#else
		exit(1);
#endif /* linux */

	}
#endif /* !darwin */
#endif /* 1 */

#if 0
#ifdef SYS_DARWIN
	/* Grab it like a keyboard interface, interface 0 */
	if (usb_claim_interface(wispy, 0) < 0) {
		printf("Unable to claim interface 0 -- are you root?\n");
		exit(1);
	}
#endif
#endif

	initscr();
	cbreak();
	noecho();

	if (COLS < WISPY_SAMPLES + 8) {
		printf("Terminal must be at least %d characters wide\n", WISPY_SAMPLES + 8);
		return 1;
	}

	if (LINES < 24) {
		printf("Terminal must be at least %d characters high\n", 24);
		return 1;
	}

	window = subwin(stdscr, LINES - 2, WISPY_SAMPLES + 2, 0, 0);
	sigwin = subwin(stdscr, LINES - 2, COLS - (WISPY_SAMPLES + 2), 0, WISPY_SAMPLES + 2);
	chwin = subwin(stdscr, 2, WISPY_SAMPLES + 2, LINES - 2, 0);
	box(window, 0, 0);

	/* Draw the signal levels for the sig window */
	mvwaddstr(sigwin, 1, 0, " 0 dBm");
	mvwaddstr(sigwin, ((LINES - 2) / 4), 0, "-15dBm");
	mvwaddstr(sigwin, ((LINES - 2) / 2), 0, "-30dBm");
	mvwaddstr(sigwin, ((LINES - 2) / 4) * 3 + 1, 0, "-45dBm");
	mvwaddstr(sigwin, (LINES - 3), 0, "-60dBm");

	/* Draw the channels */
	memset(line, ' ', WISPY_SAMPLES);
	line[WISPY_SAMPLES] = 0;
	/* First line of channels */
	line[11] = '1'; line[16] = '2';
	line[21] = '3'; line[26] = '4';
	line[31] = '5'; line[36] = '6';
	line[41] = '7'; line[46] = '8';
	line[51] = '9'; line[56] = '1'; // 10
	line[61] = '1'; line[66] = '1'; // 11 12
	line[71] = '1'; line[83] = '1'; // 13 14
	mvwaddstr(chwin, 0, 0, line);

	/* Second line of channels */
	memset(line, ' ', WISPY_SAMPLES);
	line[WISPY_SAMPLES] = 0;
	line[56] = '0'; // 10
	line[61] = '1'; line[66] = '2'; // 11 12
	line[71] = '3'; line[83] = '4'; // 13 14
	mvwaddstr(chwin, 1, 0, line);

	wrefresh(window);
	wrefresh(chwin);
	wrefresh(sigwin);
	refresh();

	while (1) {
		/* Grab a report */
		if (usb_control_msg(wispy, USB_ENDPOINT_IN + USB_TYPE_CLASS + USB_RECIP_INTERFACE,
							HID_GET_REPORT, (HID_RT_FEATURE << 8),
							0, buf, 8, TIMEOUT) == 0) {
			printf("Failure in control_msg\n");
			return 1;
		}

		/* put it into the sample records */
		for (x = 1; x < 8; x++) {
			float dbm = ((float) buf[x] * 0.95f) - 61;	
			int pos = buf[0] + x - 1;

			lastsignal[pos] = (int) dbm;
			if ((int) dbm > peaksignal[pos] || peaksignal[pos] == 0) {
				peaksignal[pos] = (int) dbm;
			}
			if ((int) dbm > lastpeaksignal[pos] || lastpeaksignal[pos] == 0) {
				lastpeaksignal[pos] = (int) dbm;
			}
			avgsignal[avgpos][pos] = (int) dbm;
		}
	
		// Loop to the next signal average on a return to 0
		if (buf[0] == 0) {
			avgpos++;
			if (avgpos >= WISPY_AVGNUM)
				avgpos = 0;
		}


		/* draw the interface */
		cursesdraw++;
		if (cursesdraw > 50) {
			int zed;
			
			for (zed = 0; zed < (GRAPH_HEIGHT); zed++) {
				memset(line, ' ', WISPY_SAMPLES);
				line[WISPY_SAMPLES] = 0;

				/* modify the line */
				for (x = 0; x < WISPY_SAMPLES; x++) {
					float graphperc;
					int a, as = 0;

					graphperc = (GRAPH_HEIGHT) * (float) ((float) abs(peaksignal[x]) / 
													   (float) abs(WISPY_MINSIG));
					if (graphperc <= zed && graphperc != 0)
						line[x] = ':';

					graphperc = (LINES - 2) * (float) ((float) abs(lastpeaksignal[x]) / 
													   (float) abs(WISPY_MINSIG));

#if 0
					for (a = 0; a < WISPY_AVGNUM; a++) {
						as += avgsignal[a][x];
					}
					as = as / WISPY_AVGNUM;

					graphperc = (GRAPH_HEIGHT) * (float) ((float) abs(as) /
													   (float) abs(WISPY_MINSIG));
#endif

#if 0
					graphperc = (LINES - 2) * (float) ((float) abs(lastsignal[x]) / 
													   (float) abs(WISPY_MINSIG));
#endif
					if (graphperc <= zed && graphperc != 0)
						line[x] = '*';
				}

				mvwaddstr(window, zed + 1, 1, line);
			}

			for (x = 0; x < WISPY_SAMPLES; x++) {
				lastpeaksignal[x] = WISPY_MINSIG;
			}

			cursesdraw = 0;
			wrefresh(window);
			refresh();
		}

		tm.tv_sec = 0;
		tm.tv_usec = 80;

		FD_ZERO(&readset);
		FD_SET(fileno(stdin), &readset);

		select(fileno(stdin)+1, &readset, NULL, NULL, &tm);

		if (FD_ISSET(fileno(stdin), &readset)) {
			char c = getch();

			if (c == 'c') {
				for (x = 0; x < WISPY_SAMPLES; x++) {
					peaksignal[x] = WISPY_MINSIG;
				}
			} else if (c == 'q') {
				break;
			}
		}
	}

	endwin();
	usb_close(wispy);

	return 0;
}	

