/* Metageek WiSPY LibUSB interface 
 * Mike Kershaw/Dragorn <dragorn@kismetwireless.net>
 *
 * LibUSB userspace control for the Metageek WiSPY spectrum analyzer
 * http://www.metageek.net/
 *
 * GTK graphical display of WiSPY spectrum analyzer
 *
 * Yes, all things considered, this is some pretty ugly code, but the app
 * isn't THAT large that it warrants a total rewrite
 *
 * This code is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Extra thanks to Ryan Woodings @ Metageek for interface documentation
 */

#include <stdio.h>
#include <usb.h>
#include <gtk/gtk.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <signal.h>

#include "config.h"

#ifdef HAVE_LINUX_DEVDISCONNECT
/* Needed for our own tools */
#include <unistd.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <linux/usbdevice_fs.h>
#include <errno.h>
#endif

/* ------------------------------------------------------------------------- */

/* First-gen firmware */
#define METAGEEK_WISPY1_VID	0x04b4
#define METAGEEK_WISPY1_PID	0x0bad

/* Second-gen firmware */
#define METAGEEK_WISPY2_VID	0x1781
#define METAGEEK_WISPY2_PID	0x083e

#define TIMEOUT 3000

/* Hid get report from spec */
#define HID_GET_REPORT 0x01
#define HID_RT_FEATURE 0x03

/* Number of samples we get from the device */
#define NUM_SAMPLES		84
/* Minimum signal level the device will report */
#define MIN_SAMPLE		-65
/* Calibration threshold for smoothing noisy units */
#define CAL_NOISE_THRESHOLD	-54
/* Maximum calibration value we accept, no matter what the threshold */
#define MAX_CAL_VALUE	4
/* Min ampl from radio */
#define MIN_SIG_AMPL	-61
/* Number of full-spectrum samples to calibrate on */
#define CAL_NUM_SAMPLES	20
/* Number of spectrum samples to keep backlogged */
#define NUM_SPECTRUM	80
/* Number of samples MAXd per drawing sample */
#define NUM_SPECTRUM_SAMPLES	12

/* channels ... mhz offset from 2400, -1 */
const int nchannels = 14;
const int choffst[] = { 11, 16, 21, 26, 31, 36, 41, 46, 51, 56, 61, 66, 71, 83 };
/* how many mhz a channel spreads out from its center point when we hilight it */
const int chanspread = 12;

/* RGB color map for spectrogram */
const char *spectromap[] = {
	"#0045FE", "#005CFE", "#0073FE", "#008AFE",
	"#00A1FE", "#00B8FE", "#00CFFE", "#00E6FE", 
	"#1AEBCB", "#33F098", "#4DF566", "#66FA33", 
	"#80FF00", "#99F500", "#B3EA00", "#CCE000", 
	"#E6D500", "#FFCB00", "#FFBC00", "#FFAD00", 
	"#FF9E00", "#FF8F00", "#FF8000", "#FF740E", 
	"#FF681C", "#FF5D29", "#FF5137", "#FF4545", 
	"#FF3953", "#FF2D61", "#FF2565"
};

/* Horizontal lines in the graph */
const int hlines = 8;
const int base_db_offset = -30;

#define GRAPH_XOFF		30
#define GRAPH_YOFF		30

#define SPECT_XOFF		30
#define SPECT_YOFF		30

/* ------------------------------------------------------------------------- */

/* IPC pid */
pid_t ipcpid;

int calibrated = 0;
int calibration_offset = 0;

/* Sample records */
int peak_samples[NUM_SAMPLES];
int last_samples[NUM_SAMPLES];
long long avg_samples[NUM_SAMPLES];
long long avg_sample_count = 0;

/* Spectrum samples */
int spectrum_samples[NUM_SPECTRUM][NUM_SAMPLES];
/* Position in the spectrum max count */
int spectrum_sample_count = 0;
/* Position in the sample ring */
int spectrum_ring_pos = 0;

/* ------------------------------------------------------------------------- */

/* Drawing mode -- 0 = graph, 1 = spectrum */
int graph_mode = 0;

/* Graph background, peak, and current signal contexts */
GdkGC *g_bg1_gc = NULL;
GdkGC *g_bg2_gc = NULL;
GdkGC *g_dln_gc = NULL;
GdkGC *g_ln_gc = NULL;
GdkGC *g_pk_gc = NULL;
GdkGC *g_cs_gc = NULL;
GdkGC *g_av_gc = NULL;
/* Red power line context */
GdkGC *g_rpw_gc = NULL;
/* Hilighted channel color */
GdkGC *g_hc_gc = NULL;

/* Graph fonts */
GdkFont *graph_font = NULL;

/* Allocated colors for spectrogram */
GdkGC *spectrogc[31];

/* ------------------------------------------------------------------------- */

/* GTK widgets we care about */
GtkWidget *gtbutton = NULL;
GtkWidget *stbutton = NULL;
GtkWidget *clearbutton = NULL;

/* Our pixmap backing for the graph itself */
GdkPixmap *pixmap = NULL;
GtkWidget *graph = NULL;

/* width and height of graph drawing area as of the last draw */
int graphx = 0, graphy = 0;

/* Coordinate pairs of the channels */
GdkPoint *chpoints;

/* Where we are for power on the graph widget (mouse) */
int pow_pos = 0;

/* Did we hilight a channel? */
int hi_chan = -1;

/* is the mouse down during an event? */
int mouse_down = 0;

/* ------------------------------------------------------------------------- */

/* Draw the graph itself */
void graph_pixmap_draw() {
	GdkRectangle update_rect;
	int x, wbar;
	/* Current points */
	GdkPoint points[NUM_SAMPLES];
	/* peak points */
	GdkPoint ppoints[NUM_SAMPLES + 2];
	/* avg points */
	GdkPoint apoints[NUM_SAMPLES + 2];

	/* height of text */
	int texth, textw;
	char text[8];

	/* We're going to redraw everything */
	update_rect.x = 0;
	update_rect.y = 0;
	update_rect.width = graph->allocation.width;
	update_rect.height = graph->allocation.height;

	/* erase it */
	gdk_draw_rectangle(pixmap, g_bg1_gc, TRUE, 
					   0, 0, graph->allocation.width, graph->allocation.height);

	/* left buffer, 80 right buffer for font */
	graphx = graph->allocation.width - GRAPH_XOFF - 80;

	/* then adjust to an even multiple of num_samples */
	wbar = graphx / (NUM_SAMPLES - 1);

	/* rescale the graph to an even multiple */
	graphx = wbar * (NUM_SAMPLES - 1);

	/* top, 40 bottom for font */
	graphy = graph->allocation.height - GRAPH_YOFF - 40;

	/* Plot the lighter rectangle */
	gdk_draw_rectangle(pixmap, g_bg2_gc, TRUE, GRAPH_XOFF, GRAPH_YOFF, graphx, graphy);

	/* Plot the channel hilight if we need to, and draw the "you picked" text */
	if (hi_chan >= 0) {
		int center, spread;
		char mtext[128];
		int start, end, smod;

		/* Plot the "you have picked" text */
		snprintf(mtext, 128, "SELECTED: %d, 24%02dMHz",
				 hi_chan + 1, choffst[hi_chan] + 1);

		texth = gdk_string_height(graph_font, mtext);
		textw = gdk_string_width(graph_font, mtext);
		gdk_draw_string(pixmap, graph_font, g_ln_gc,
						GRAPH_XOFF + graphx - textw, 
						GRAPH_YOFF - texth,
						mtext);

		/* Fit the hilight within our graph */
		smod = 0;
		center = GRAPH_XOFF + (choffst[hi_chan] * wbar);
		spread = (chanspread * wbar);

		start = center - spread;
		if (start < GRAPH_XOFF) {
			smod = GRAPH_XOFF - start;
			start = GRAPH_XOFF;
		}

		end = spread * 2 - smod;
		if (start + end > GRAPH_XOFF + graphx)
			end = (GRAPH_XOFF + graphx) - start;

		/* plot the really light rectangle */
		gdk_draw_rectangle(pixmap, g_hc_gc, TRUE,
						   start, GRAPH_YOFF,
						   end, graphy);
	}

	/* Draw the dBm lines and power labels */
	for (x = 0; x <= (hlines - 2) && calibrated; x++) {
		int dbm, py;
		dbm = base_db_offset + ((x) * -5);
		py = (float) graphy * (float) ((float) (abs(dbm) + base_db_offset) / 
									   (float) (abs(MIN_SAMPLE) + base_db_offset));
		gdk_draw_line(pixmap, g_dln_gc, GRAPH_XOFF, GRAPH_YOFF + py,
					  GRAPH_XOFF + graphx, GRAPH_YOFF + py);

		snprintf(text, 8, "%d dBm", dbm);
		texth = gdk_string_height(graph_font, text);
		gdk_draw_string(pixmap, graph_font, g_ln_gc,
						GRAPH_XOFF + graphx + 5, 
						GRAPH_YOFF + py + (texth / 2),
						text);
	}

	/* Draw the channel text along the bottom */
	for (x = 0; x < nchannels; x++) {
		GdkGC *gc = g_ln_gc;

		if (x == hi_chan) {
			gc = g_cs_gc;
			snprintf(text, 8, "<%d>", x + 1);
		} else {
			snprintf(text, 8, "%d", x + 1);
		}

		gdk_draw_line(pixmap, g_ln_gc,
					  GRAPH_XOFF + (choffst[x] * wbar), GRAPH_YOFF + graphy,
					  GRAPH_XOFF + (choffst[x] * wbar), GRAPH_YOFF + graphy + 3);

		texth = gdk_string_height(graph_font, text);
		textw = gdk_string_width(graph_font, text);
		gdk_draw_string(pixmap, graph_font, gc, 
						GRAPH_XOFF + (choffst[x] * wbar) - textw / 2, 
						GRAPH_YOFF + graphy + 10 + texth, 
						text);

		chpoints[x].x = GRAPH_XOFF + (choffst[x] * wbar) - textw / 2 - 4;
		chpoints[x].y = GRAPH_YOFF + graphy + 10 - 4;

		chpoints[x + nchannels].x = chpoints[x].x + textw + 8;
		chpoints[x + nchannels].y = chpoints[x].y + texth + 10;
	}

	/* Bail if we're not calibrated yet */
	if (calibrated == 0) {
		/* rectangle */
		gdk_draw_rectangle(pixmap, g_ln_gc, 0, GRAPH_XOFF, GRAPH_YOFF, graphx, graphy);

		texth = gdk_string_height(graph_font, "Calibrating...");
		textw = gdk_string_width(graph_font, "Calibrating...");
		gdk_draw_string(pixmap, graph_font, g_ln_gc,
						GRAPH_XOFF + (graphx / 2) - (textw / 2), 
						GRAPH_YOFF + (graphy / 2) - (texth / 2),
						"Calibrating...");

		/* Mark that we need to update */
		gtk_widget_draw(graph, &update_rect);

		return;
	}

	/* First poly point: lower left */
	ppoints[0].x = GRAPH_XOFF;
	ppoints[0].y = GRAPH_YOFF + graphy;
	apoints[0].x = GRAPH_XOFF;
	apoints[0].y = GRAPH_YOFF + graphy;

	for (x = 0; x < NUM_SAMPLES; x++) {
		int px, py;

		px = GRAPH_XOFF + (x * wbar);

		/* Current scaled */
		py = (float) graphy * (float) ((float) (abs(last_samples[x]) + base_db_offset) / 
									   (float) (abs(MIN_SAMPLE) + base_db_offset));
		points[x].x = px;
		if (py > graphy)
			py = graphy;
		points[x].y = py + GRAPH_YOFF;

		/* peak scaled */
		py = (float) graphy * (float) ((float) (abs(peak_samples[x]) + base_db_offset) / 
									   (float) (abs(MIN_SAMPLE) + base_db_offset));

		ppoints[x + 1].x = px;
		if (py > graphy)
			py = graphy;
		ppoints[x + 1].y = py + GRAPH_YOFF;

		/* avg */
		if (avg_sample_count == 0 || avg_samples[x] == 0)
			py = graphy;
		else
			py = (float) graphy * 
				(float) ((float) (abs(avg_samples[x] / avg_sample_count) + base_db_offset) /
						 (float) (abs(MIN_SAMPLE) + base_db_offset));
		apoints[x + 1].x = px;
		if (py > graphy)
			py = graphy;
		apoints[x + 1].y = py + GRAPH_YOFF;
	}

	/* Last poly point:  lower right */
	ppoints[NUM_SAMPLES + 1].x = GRAPH_XOFF + graphx;
	ppoints[NUM_SAMPLES + 1].y = GRAPH_YOFF + graphy;
	apoints[NUM_SAMPLES + 1].x = GRAPH_XOFF + graphx;
	apoints[NUM_SAMPLES + 1].y = GRAPH_YOFF + graphy;

	/* Peak dark polygon */
	gdk_draw_polygon(pixmap, g_pk_gc, TRUE, ppoints, NUM_SAMPLES + 2);

	/* Avg green polygon */
	gdk_draw_polygon(pixmap, g_av_gc, TRUE, apoints, NUM_SAMPLES + 2);

	/* current yellow line */
	gdk_draw_lines(pixmap, g_cs_gc, points, NUM_SAMPLES);

	/* Finish the channel hilighting */
	if (hi_chan >= 0) {
		int center, spread, start, end, smod;

		smod = 0;
		center = GRAPH_XOFF + (choffst[hi_chan] * wbar);
		spread = (chanspread * wbar);

		center = GRAPH_XOFF + (choffst[hi_chan] * wbar);
		spread = (chanspread * wbar);

		start = center - spread;
		if (start < GRAPH_XOFF) {
			smod = GRAPH_XOFF - start;
			start = GRAPH_XOFF;
		}

		end = spread * 2 - smod;
		if (start + end > GRAPH_XOFF + graphx)
			end = (GRAPH_XOFF + graphx) - start;

		/* plot the vertical dashed bars on top */
		gdk_draw_line(pixmap, g_dln_gc, start, GRAPH_YOFF,
					  start, GRAPH_YOFF + graphy);
		gdk_draw_line(pixmap, g_dln_gc, start + end, GRAPH_YOFF,
					  start + end, GRAPH_YOFF + graphy);
		/* plot the vertical indicator */
		gdk_draw_line(pixmap, g_ln_gc, center, GRAPH_YOFF,
					  center, GRAPH_YOFF + graphy);

		/* draw the thickened top and bottom bars */
		gdk_draw_rectangle(pixmap, g_ln_gc, 1, start, GRAPH_YOFF + 1, end, 3);
		gdk_draw_rectangle(pixmap, g_ln_gc, 1, start, GRAPH_YOFF + graphy + 1, end, 3);
	}

	/* Power mouse position */
	if (pow_pos != 0) {
		int py;
		py = (float) graphy * (float) ((float) (abs(pow_pos) + base_db_offset) / 
									   (float) (abs(MIN_SAMPLE) + base_db_offset));
		gdk_draw_line(pixmap, g_rpw_gc, GRAPH_XOFF, GRAPH_YOFF + py, 
					  GRAPH_XOFF + graphx, GRAPH_YOFF + py);

		/* Red rectangle surrounded by white line */
		gdk_draw_rectangle(pixmap, g_rpw_gc, 1, 
						   GRAPH_XOFF - 15, GRAPH_YOFF + py - 5, 10, 10);
		gdk_draw_rectangle(pixmap, g_ln_gc, 0, 
						   GRAPH_XOFF - 15, GRAPH_YOFF + py - 5, 10, 10);
		/* Draw the text */
		snprintf(text, 8, "%d", pow_pos);
		textw = gdk_string_width(graph_font, text);
		texth = gdk_string_height(graph_font, text);
		gdk_draw_string(pixmap, graph_font, g_ln_gc,
						GRAPH_XOFF - textw - 5, 
						GRAPH_YOFF + py + 5 + 4 + texth,
						text);
	}

	/* rectangle */
	gdk_draw_rectangle(pixmap, g_ln_gc, 0, GRAPH_XOFF, GRAPH_YOFF, graphx, graphy);

	/* Mark that we need to update */
	gtk_widget_draw(graph, &update_rect);
}

/* Draw the spectrum */
void spectrum_pixmap_draw() {
	GdkRectangle update_rect;
	int x, y, sp, wbar, vbar, spec_pos;

	/* height of text */
	int texth, textw;
	char text[8];

	/* We're going to redraw everything */
	update_rect.x = 0;
	update_rect.y = 0;
	update_rect.width = graph->allocation.width;
	update_rect.height = graph->allocation.height;

	/* erase it */
	gdk_draw_rectangle(pixmap, g_bg1_gc, TRUE, 
					   0, 0, graph->allocation.width, graph->allocation.height);

	/* left buffer, 20 right buffer for font */
	graphx = graph->allocation.width - SPECT_XOFF - 20;

	/* then adjust to an even multiple of num_samples */
	wbar = graphx / (NUM_SAMPLES);

	/* rescale the graph to an even multiple */
	graphx = wbar * (NUM_SAMPLES);

	/* top, 20 bottom for font */
	graphy = graph->allocation.height - SPECT_YOFF - 20;

	/* Adjust to an even multiple of spectrum samples */
	vbar = graphy / (NUM_SPECTRUM);

	/* rescale the graph to an even multiple */
	graphy = vbar * (NUM_SPECTRUM);

	/* Plot the lighter rectangle */
	gdk_draw_rectangle(pixmap, g_bg2_gc, TRUE, SPECT_XOFF, SPECT_YOFF, graphx, graphy);

	/* Draw the channel text along the bottom */
	for (x = 0; x < nchannels; x++) {
		GdkGC *gc = g_ln_gc;

		if (x == hi_chan) {
			gc = g_cs_gc;
			snprintf(text, 8, "<%d>", x + 1);
		} else {
			snprintf(text, 8, "%d", x + 1);
		}

		gdk_draw_line(pixmap, g_ln_gc,
					  GRAPH_XOFF + (choffst[x] * wbar), GRAPH_YOFF + graphy,
					  GRAPH_XOFF + (choffst[x] * wbar), GRAPH_YOFF + graphy + 3);

		texth = gdk_string_height(graph_font, text);
		textw = gdk_string_width(graph_font, text);
		gdk_draw_string(pixmap, graph_font, gc, 
						GRAPH_XOFF + (choffst[x] * wbar) - textw / 2, 
						GRAPH_YOFF + graphy + 10 + texth, 
						text);

		chpoints[x].x = GRAPH_XOFF + (choffst[x] * wbar) - textw / 2 - 4;
		chpoints[x].y = GRAPH_YOFF + graphy + 10 - 4;

		chpoints[x + nchannels].x = chpoints[x].x + textw + 8;
		chpoints[x + nchannels].y = chpoints[x].y + texth + 10;
	}

#if 0
	/* Draw the y axis label on the right */
	DrawVertText(pixmap, g_ln_gc, graph_font, 5, 20 + graphx + 10, graphy / 2, 
				 "Time Elapsed", 4);
#endif

	/* Bail if we're not calibrated yet */
	if (calibrated == 0) {
		/* rectangle */
		gdk_draw_rectangle(pixmap, g_ln_gc, 0, SPECT_XOFF, SPECT_YOFF, graphx, graphy);

		texth = gdk_string_height(graph_font, "Calibrating...");
		textw = gdk_string_width(graph_font, "Calibrating...");
		gdk_draw_string(pixmap, graph_font, g_ln_gc,
						SPECT_XOFF + (graphx / 2) - (textw / 2), 
						SPECT_YOFF + (graphy / 2) - (texth / 2),
						"Calibrating...");

		/* Mark that we need to update */
		gtk_widget_draw(graph, &update_rect);

		return;
	}

	/* Loop the ring buffer */
	y = spectrum_ring_pos - 1;
	/* Local pos is 0 */
	sp = 0;
	while (1) {
		int py, px;
		GdkGC *cgc;

		/* Set the spectrum position (another ring-aware) */
		if (y < 0)
			spec_pos = (NUM_SPECTRUM - 1);
		else
			spec_pos = y;

		py = SPECT_YOFF + ((NUM_SPECTRUM - sp - 1) * vbar);

		for (x = 0; x < NUM_SAMPLES; x++) {
			int samp;

			px = SPECT_XOFF + (x * wbar);

			/* Figure out the color */
			samp = spectrum_samples[spec_pos][x] - MIN_SIG_AMPL;

			if (samp < 0) {
				samp = 0;
			} else if (samp > 30) {
				samp = 30;
			}

			cgc = spectrogc[samp];

			gdk_draw_rectangle(pixmap, cgc, 1, px, py, wbar, vbar);
		}

		/* Loop our local point */
		sp++;

		/* Loop down the ringbuffer */
		y--;
		if (y < -1)
			y = NUM_SPECTRUM - 1;

		if (y == spectrum_ring_pos)
			break;
	}

	/* Finish the channel hilighting */
	if (hi_chan >= 0) {
		int center, spread, start, end, smod;
		char mtext[128];

		/* Plot the "you have picked" text */
		snprintf(mtext, 128, "SELECTED: %d, 24%02dMHz",
				 hi_chan + 1, choffst[hi_chan] + 1);

		texth = gdk_string_height(graph_font, mtext);
		textw = gdk_string_width(graph_font, mtext);
		gdk_draw_string(pixmap, graph_font, g_ln_gc,
						GRAPH_XOFF + graphx - textw, 
						GRAPH_YOFF - texth,
						mtext);

		smod = 0;
		center = GRAPH_XOFF + (choffst[hi_chan] * wbar);
		spread = (chanspread * wbar);

		center = GRAPH_XOFF + (choffst[hi_chan] * wbar);
		spread = (chanspread * wbar);

		start = center - spread;
		if (start < GRAPH_XOFF) {
			smod = GRAPH_XOFF - start;
			start = GRAPH_XOFF;
		}

		end = spread * 2 - smod;
		if (start + end > GRAPH_XOFF + graphx)
			end = (GRAPH_XOFF + graphx) - start;

		/* plot the vertical dashed bars on top */
		gdk_draw_line(pixmap, g_dln_gc, start, GRAPH_YOFF,
					  start, GRAPH_YOFF + graphy);
		gdk_draw_line(pixmap, g_dln_gc, start + end, GRAPH_YOFF,
					  start + end, GRAPH_YOFF + graphy);
		/* plot the vertical indicator */
		gdk_draw_line(pixmap, g_ln_gc, center, GRAPH_YOFF,
					  center, GRAPH_YOFF + graphy);

		/* draw the thickened top and bottom bars */
		gdk_draw_rectangle(pixmap, g_ln_gc, 1, start, GRAPH_YOFF + 1, end, 3);
		gdk_draw_rectangle(pixmap, g_ln_gc, 1, start, GRAPH_YOFF + graphy + 1, end, 3);
	}

	/* rectangle */
	gdk_draw_rectangle(pixmap, g_ln_gc, 0, SPECT_XOFF, SPECT_YOFF, graphx, graphy);

	/* Mark that we need to update */
	gtk_widget_draw(graph, &update_rect);
}

/* Configure event:  Build a graph pixbuf */
static gint graph_configure_event(GtkWidget *widget, GdkEventConfigure *event) {
	GdkColormap *colormap;
	GdkColor color;
	int x;

	if (pixmap)
		gdk_pixmap_unref(pixmap);

	/* Size of the parent widget, same depth */
	pixmap = gdk_pixmap_new(widget->window, widget->allocation.width,
							widget->allocation.height, -1);

	/* Get the system color map */
	colormap = gdk_colormap_get_system();

	if (g_bg1_gc == NULL) {
		gdk_color_parse("#191970", &color);
		gdk_colormap_alloc_color(colormap, &color, FALSE, TRUE);

		g_bg1_gc = gdk_gc_new(pixmap);
		gdk_gc_set_foreground(g_bg1_gc, &color);
	}

	if (g_bg2_gc == NULL) {
		gdk_color_parse("#00008B", &color);
		gdk_colormap_alloc_color(colormap, &color, FALSE, TRUE);

		g_bg2_gc = gdk_gc_new(pixmap);
		gdk_gc_set_foreground(g_bg2_gc, &color);
	}

	if (g_hc_gc == NULL) {
		gdk_color_parse("#493EB9", &color);
		gdk_colormap_alloc_color(colormap, &color, FALSE, TRUE);

		g_hc_gc = gdk_gc_new(pixmap);
		gdk_gc_set_foreground(g_hc_gc, &color);
	}

	if (g_pk_gc == NULL) {
		gdk_color_parse("black", &color);
		gdk_colormap_alloc_color(colormap, &color, FALSE, TRUE);

		g_pk_gc = gdk_gc_new(pixmap);
		gdk_gc_set_foreground(g_pk_gc, &color);
	}

	if (g_av_gc == NULL) {
		gdk_color_parse("#2E8B57", &color);
		gdk_colormap_alloc_color(colormap, &color, FALSE, TRUE);

		g_av_gc = gdk_gc_new(pixmap);
		gdk_gc_set_foreground(g_av_gc, &color);
	}

	if (g_cs_gc == NULL) {
		gdk_color_parse("yellow", &color);
		gdk_colormap_alloc_color(colormap, &color, FALSE, TRUE);

		g_cs_gc = gdk_gc_new(pixmap);
		gdk_gc_set_foreground(g_cs_gc, &color);
		gdk_gc_set_line_attributes(g_cs_gc, 2, GDK_LINE_SOLID, 
								   GDK_CAP_BUTT, GDK_JOIN_MITER);
	}

	if (g_dln_gc == NULL) {
		gchar list[] = {1, 1};

		gdk_color_parse("white", &color);
		gdk_colormap_alloc_color(colormap, &color, FALSE, TRUE);

		g_dln_gc = gdk_gc_new(pixmap);
		gdk_gc_set_foreground(g_dln_gc, &color);

		gdk_gc_set_dashes(g_dln_gc, 0, list, sizeof(list));
		gdk_gc_set_line_attributes(g_dln_gc, 0, GDK_LINE_ON_OFF_DASH, 
								   GDK_CAP_BUTT, GDK_JOIN_MITER);
	}

	if (g_rpw_gc == NULL) {
		gchar list[] = {4, 2};

		gdk_color_parse("red", &color);
		gdk_colormap_alloc_color(colormap, &color, FALSE, TRUE);

		g_rpw_gc = gdk_gc_new(pixmap);
		gdk_gc_set_foreground(g_rpw_gc, &color);

		gdk_gc_set_dashes(g_rpw_gc, 0, list, sizeof(list));
		gdk_gc_set_line_attributes(g_rpw_gc, 0, GDK_LINE_ON_OFF_DASH, 
								   GDK_CAP_BUTT, GDK_JOIN_MITER);
	}

	if (g_ln_gc == NULL) {
		gdk_color_parse("white", &color);
		gdk_colormap_alloc_color(colormap, &color, FALSE, TRUE);

		g_ln_gc = gdk_gc_new(pixmap);
		gdk_gc_set_foreground(g_ln_gc, &color);

		gdk_gc_set_line_attributes(g_ln_gc, 0, GDK_LINE_SOLID, 
								   GDK_CAP_BUTT, GDK_JOIN_MITER);
	}

	if (spectrogc[0] == NULL) {
		for (x = 0; x < 31; x++) {
			gdk_color_parse(spectromap[x], &color);
			gdk_colormap_alloc_color(colormap, &color, FALSE, TRUE);

			spectrogc[x] = gdk_gc_new(pixmap);
			gdk_gc_set_foreground(spectrogc[x], &color);
		}
	}

	if (graph_font == NULL) {
		graph_font = gdk_font_load("-*-helvetica-*-r-*-*-12-*-*-*-*-*-*-*");
	}

	/* redraw it from data */
	if (graph_mode == 0) {
		graph_pixmap_draw();
	} else {
		spectrum_pixmap_draw();
	}

	return TRUE;
}

/* Expose event:  copy the pixbuf back to the screen */
static gint graph_expose_event(GtkWidget *widget, GdkEventExpose *event) {
	gdk_draw_pixmap(widget->window, widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
					pixmap, event->area.x, event->area.y,
					event->area.x, event->area.y,
					event->area.width, event->area.height);
	return FALSE;
}

/* Mouse button event -- if we're inside the graph itself, 
 * figure out the power and start drawing the red line */
static gint graph_button_press_event(GtkWidget *widget, GdkEventButton *event) {
	int dbm;
	int maxy;

	if (event->button != 1 || pixmap == NULL || graphy == 0)
		return TRUE;

	/* Search the channel positions, if we can... hijack dbm int for our
	 * loop control */
	maxy = 0;
	for (dbm = 0; dbm < nchannels; dbm++) {
		// Remember the tallest letter
		if (chpoints[dbm + nchannels].y > maxy)
			maxy = chpoints[dbm + nchannels].y;

		if (event->x > chpoints[dbm].x && event->x < chpoints[dbm + nchannels].x &&
			event->y > chpoints[dbm].y && event->y < chpoints[dbm + nchannels].y) {
			hi_chan = dbm;
			return TRUE;
		}

		// If we didn't hit a label but we're within the label bar, then clear
		// the channel selection
		if ((event->x > GRAPH_XOFF && event->x < GRAPH_XOFF + graphx) &&
			(event->y > GRAPH_YOFF + graphy && event->y < maxy + 5)) {
			hi_chan = -1;
		}
	}

	/* Bail if we're not inside the graph area, and we don't do powpos for spectro */
	if (graph_mode != 0 || event->x < GRAPH_XOFF || event->y < GRAPH_YOFF ||
		event->x > GRAPH_XOFF + graphx || event->y > GRAPH_YOFF + graphy)
		return TRUE;

	/* calculate the dbm at this pos */
	dbm = (float) (abs(MIN_SAMPLE) + base_db_offset) *
		((float) ((event->y - GRAPH_YOFF) / graphy));

	pow_pos = (dbm * -1) + base_db_offset;

	return TRUE;
}

/* Mouse move event -- if we're inside the graph, allow them to drag the power
 * around */
static gint graph_motion_event(GtkWidget *widget, GdkEventMotion *event) {
	int x, y, dbm;
	GdkModifierType state;

	if (event->is_hint) {
		gdk_window_get_pointer(event->window, &x, &y, &state);
	} else {
		x = (int) event->x;
		y = (int) event->y;
		state = event->state;
	}

	if (pixmap != NULL && graph_mode == 0) {
		/* If the mouse button was down, is still down, and we're outside the
		 * area, make the power line go away */
		if (mouse_down && (state & GDK_BUTTON1_MASK) &&
			(x < GRAPH_XOFF || y < GRAPH_YOFF || 
			 x > GRAPH_XOFF + graphx || y > GRAPH_YOFF + graphy)) {
			pow_pos = 0;
			mouse_down = 0;
			return TRUE;
		}

		/* Bail if we're not inside the graph area */
		if (x < GRAPH_XOFF || y < GRAPH_YOFF || 
			x > GRAPH_XOFF + graphx || y > GRAPH_YOFF + graphy) {
			mouse_down = 0;
			return TRUE;
		}

		/* If the mouse is down, move */
		if ((state & GDK_BUTTON1_MASK)) {
			mouse_down = 1;
			/* calculate the dbm at this pos */
			dbm = (float) (abs(MIN_SAMPLE) + base_db_offset) *
				((float) ((float) (y - GRAPH_YOFF) / graphy));

			pow_pos = (dbm * -1) + base_db_offset;
			return TRUE;
		}

		/* let go of the mouse at the end */
		mouse_down = 0;
	}

	return TRUE;
}

void graph_toggle_callback(GtkWidget *widget, gpointer data) {
	// Grab the button
	if (widget == gtbutton) {
		if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget))) {
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(data), 0);
			graph_mode = 0;
		} else {
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(data), 1);
			graph_mode = 1;
		}

		return;
	}

	if (widget == stbutton) {
		if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget))) {
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(data), 0);
			graph_mode = 1;
		} else {
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(data), 1);
			graph_mode = 0;
		}

		return;
	}
}

void clearcallback(GtkWidget *widget, gpointer data) {
	int x,y;
	for (x = 0; x < NUM_SAMPLES; x++) {
		peak_samples[x] = MIN_SAMPLE;
		avg_samples[x] = 0;
		avg_sample_count = 0;
		for (y = 0; y < NUM_SPECTRUM; y++) {
			spectrum_samples[y][x] = MIN_SAMPLE;
		}
	}
}

gint delete_event(GtkWidget *widget, GdkEvent *event,
				  gpointer data) {
	gtk_main_quit();
	kill(ipcpid, SIGKILL);
	return(FALSE);
}

/* Basic device initialize, grabs either version 1 or version 2 */
static struct usb_device *init_wispy_device() {
	struct usb_bus *bus;
	struct usb_device *dev;

	/* Libusb init */
	usb_init();
	usb_find_busses();
	usb_find_devices();

	/* Scan all the devices on all the busses */
	for (bus = usb_busses; bus; bus = bus->next) {
		for (dev = bus->devices; dev; dev = dev->next) {
			if (((dev->descriptor.idVendor == METAGEEK_WISPY1_VID) &&
				 (dev->descriptor.idProduct == METAGEEK_WISPY1_PID)) ||
				((dev->descriptor.idVendor == METAGEEK_WISPY2_VID) &&
				 (dev->descriptor.idProduct == METAGEEK_WISPY2_PID)))
				return dev;
		}
	}

	return NULL;
}

/* GDK callback that services the drawing engine side of the callback */
void ipc_poll_callback(gpointer data, gint source, GdkInputCondition condition) {
	char buf[8];
	int x, y;

	if (read(source, buf, 8) < 0) {
		printf("wispy_gtk: Failed to read signal data.\n");
		exit(1);
	}

	if (buf[0] == 0) {
		if (graph_mode == 0)
			graph_pixmap_draw();
		else
			spectrum_pixmap_draw();

		avg_sample_count++;

		/* Calculate the calibration if we're uncalibrated and have enough loops */
		if (avg_sample_count >= CAL_NUM_SAMPLES && calibrated == 0) {
			/* Calibration */
			long amptotal = 0;
			int calfreqs = 0;

			/* Calc the calibration offset */
			for (x = 0; x < NUM_SAMPLES; x++) {
				if (peak_samples[x] < CAL_NOISE_THRESHOLD) {
					amptotal += avg_samples[x];
					calfreqs++;
				}
			}

			if (calfreqs != 0) {
				calibration_offset = 
					(float) (((float) amptotal / (float) calfreqs) / 
							 (float) CAL_NUM_SAMPLES) - MIN_SIG_AMPL;
			}

			if (calibration_offset > MAX_CAL_VALUE)
				calibration_offset = MAX_CAL_VALUE;

			/* Zero the sample buffers */
			for (x = 0; x < NUM_SAMPLES; x++) {
				peak_samples[x] = MIN_SAMPLE;
				last_samples[x] = MIN_SAMPLE;
				avg_samples[x] = 0;

				for (y = 0; y < NUM_SPECTRUM; y++) {
					spectrum_samples[y][x] = MIN_SIG_AMPL;
				}
			}
			avg_sample_count = 0;

			calibrated = 1;
			return;
		}

		/* Increase the spectrum sample, jump to the next position in the ring,
		 * and loop the ring if we need to */
		spectrum_sample_count++;
		if (spectrum_sample_count >= NUM_SPECTRUM_SAMPLES) {
			spectrum_sample_count = 0;
			spectrum_ring_pos++;
			if (spectrum_ring_pos >= NUM_SPECTRUM)
				spectrum_ring_pos = 0;
			for (x = 0; x < NUM_SAMPLES; x++)
				spectrum_samples[spectrum_ring_pos][x] = MIN_SIG_AMPL;
		}

		for (x = 0; x < NUM_SAMPLES; x++)
			last_samples[x] = MIN_SAMPLE;
	}

	/* put it into the sample records */
	for (x = 1; x < 8; x++) {
		/* Convert the rssi */
		float dbm = ((float) buf[x] * 0.95f) - 61;	
		/* Subtract the calibration normalizer */
		dbm -= calibration_offset;
		/* check the floor */
		if (dbm < MIN_SIG_AMPL)
			dbm = MIN_SIG_AMPL;

		int pos = buf[0] + x - 1;

		/* overall peak value */
		if ((int) dbm > peak_samples[pos]) {
			peak_samples[pos] = (int) dbm;
		}

		/* peak since last draw */
		if ((int) dbm > last_samples[pos]) {
			last_samples[pos] = (int) dbm;
		}

		if ((int) dbm > spectrum_samples[spectrum_ring_pos][pos]) {
			spectrum_samples[spectrum_ring_pos][pos] = (int) dbm;
		}

		avg_samples[pos] += (int) dbm;
	}
}

/* IPC process that reads and dumps USB data to the socket pair descriptor */
void IPC_Service_Loop(int sock, struct usb_device *dev, struct usb_dev_handle *wispy) {
	char buf[8];
	struct timeval tm;
	int x;

	memset(buf, 0, 8);

	while (1) {
		/* Grab a report */
		if (usb_control_msg(wispy, USB_ENDPOINT_IN + USB_TYPE_CLASS + USB_RECIP_INTERFACE,
							HID_GET_REPORT, (HID_RT_FEATURE << 8),
							0, buf, 8, TIMEOUT) == 0) {
			printf("IPC USB poller failed on usb_control_msg\n");
			return;
		}

		/* Write it to the descriptor */
		if (write(sock, buf, 8) < 0) {
			printf("wispy_gtk: Write error on IPC to gui: %s\n", strerror(errno));
			usb_close(wispy);
			exit(1);
		}

		/*
		 * We don't actually need to do this it seems, the usb fetch and the write
		 * delays seem to throttle us down enough
		tm.tv_sec = 0;
		tm.tv_usec = 1000;
		*/

		select(0, NULL, NULL, NULL, &tm); 
	}
}

#ifdef HAVE_LINUX_DEVDISCONNECT
/* Libusb doesn't seem to always provide this, so we'll use our own, taken from the 
* usb_detatch_kernel_driver_np...
*
* THIS IS A HORRIBLE EVIL HACK THAT SHOULDN'T BE DONE, EVER
* 
*/
struct local_usb_ioctl {
	int ifno;
	int ioctl_code;
	void *data;
};

struct ghetto_libusb_devhandle {
	int fd;
	/* Nooo... so bad. */
};

int usb_detach_kernel_driver(struct usb_dev_handle *dev, int interface) {
	struct local_usb_ioctl command;
	struct ghetto_libusb_devhandle *gdev;

	command.ifno = interface;
	command.ioctl_code = USBDEVFS_DISCONNECT;
	command.data = NULL;

	gdev = (struct ghetto_libusb_devhandle *) dev;

	if (ioctl(gdev->fd, USBDEVFS_IOCTL, &command) < 0) {
		fprintf(stderr, "Could not detach kernel driver from interface %d: %s",
			interface, strerror(errno));
		return -1;
	}

	return 0;
}
#endif

int main(int argc, char *argv[]) {
	/* GTK items */
	GtkWidget *window;
	GtkWidget *packbox;
	GtkWidget *hpackbox;

	/* LibUSB items we grab before we fork */
	struct usb_device *dev;
	struct usb_dev_handle *wispy;

	/* IPC stuff */
	int sockpair[2];

	/* Junk buffer */
	char jbuf[8];

	int x, y;

	dev = init_wispy_device();
	if (!dev) {
		printf("Unable to find device on USB bus.\n");
		exit(1);
	}

	/* Open the wispy device */
	wispy = usb_open(dev);
	if (!wispy) {
		printf("Unable to open device -- are you root?\n");
		exit(1);
	}

#if 1
#ifndef SYS_DARWIN
	/* Grab it like a keyboard interface, interface 0 */
	if (usb_claim_interface(wispy, 0) < 0) {
		printf("Unable to claim interface 0 -- are you root?\n"
			   "Is something else using it (another program or the kernel HID)?\n");
#ifdef HAVE_LINUX_DEVDISCONNECT
		printf("Using terrible hack to try to detach from the Linux HID...\n"
			   "This might not work at all\n");

		if (usb_detach_kernel_driver(wispy, 0) < 0) {
			printf("Failed to detatch from HID.\n");
			exit(1);
		}

		if (usb_claim_interface(wispy, 0) < 0) {
			printf("Seemed to detach from kernel but still couldn't claim interface.\n");
			exit(1);
		}

		printf("Detatched the USB interface from the kernel driver.\n");
#else
		exit(1);
#endif

	}
#endif
#endif

#if 0
#ifndef SYS_DARWIN
	/* Grab it like a keyboard interface, interface 0 */
	if (usb_claim_interface(wispy, 0) < 0) {
		printf("Unable to claim interface 0 -- are you root? "
			   "Is something else using it (another program or the kernel HID)?\n");
		exit(1);
	}
#endif
#endif

	/* Test a read before we fork */
	if (usb_control_msg(wispy, USB_ENDPOINT_IN + USB_TYPE_CLASS + USB_RECIP_INTERFACE,
						HID_GET_REPORT, (HID_RT_FEATURE << 8),
						0, jbuf, 8, TIMEOUT) == 0) {
		printf("Couldn't poll device -- are you root?\n");
		exit(1);
	}

	if (socketpair(PF_UNIX, SOCK_DGRAM, 0, sockpair) < 0) {
		printf("Unable to create socket pair for IPC w/ USB capture process\n");
		exit(1);
	}

	if ((ipcpid = fork()) < 0) {
		printf("Unable to fork() IPC USB capture process\n");
		return 1;
	}

	if (ipcpid == 0) {
		close(sockpair[0]);
		IPC_Service_Loop(sockpair[1], dev, wispy);
		exit(0);
	}

	close(sockpair[1]);

	/* Zero the sample buffers */
	for (x = 0; x < NUM_SAMPLES; x++) {
		peak_samples[x] = MIN_SAMPLE;
		last_samples[x] = MIN_SAMPLE;
		avg_samples[x] = 0;
		for (y = 0; y < NUM_SPECTRUM; y++) {
			spectrum_samples[y][x] = MIN_SIG_AMPL;
		}
	}

	/* Zero the color map */
	spectrogc[0] = NULL;

	/* allocate the channel points */
	chpoints = (GdkPoint *) malloc(sizeof(GdkPoint) * nchannels * 2);

	/* initialize gtk */
	gtk_init(&argc, &argv);

	/* create the window */
	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);

	/* set the title */
	gtk_window_set_title(GTK_WINDOW(window), "WiSPY");

	/* set the size */
	gtk_window_set_default_size(GTK_WINDOW(window), 750, 450);

	/* set the delete event that just quits when we get a kill */
	gtk_signal_connect(GTK_OBJECT(window), "delete_event",
					   GTK_SIGNAL_FUNC(delete_event), NULL);

	/* set window border */
	gtk_container_set_border_width(GTK_CONTAINER(window), 5);

	/* make a vertical packing box */
	packbox = gtk_vbox_new(FALSE, 0);

	/* Put the box in the window */
	gtk_container_add(GTK_CONTAINER(window), packbox);

	/* Make the graph drawing area, a subsize of the available area in the window */
	graph = gtk_drawing_area_new();

	gtk_drawing_area_size(GTK_DRAWING_AREA(graph), window->allocation.width - 50,
						  window->allocation.height - 50);

	/* Attach the signals for configure and expose */
	gtk_signal_connect(GTK_OBJECT(graph), "expose_event",
					   (GtkSignalFunc) graph_expose_event, NULL);
	gtk_signal_connect(GTK_OBJECT(graph), "configure_event",
					   (GtkSignalFunc) graph_configure_event, NULL);
	/* Attach mouse */
	gtk_signal_connect(GTK_OBJECT(graph), "button_press_event",
					   (GtkSignalFunc) graph_button_press_event, NULL);
	gtk_signal_connect(GTK_OBJECT(graph), "motion_notify_event",
					   (GtkSignalFunc) graph_motion_event, NULL);

	/* Set mask */
	gtk_widget_set_events(graph, GDK_EXPOSURE_MASK |
						  GDK_LEAVE_NOTIFY_MASK |
						  GDK_BUTTON_PRESS_MASK |
						  GDK_POINTER_MOTION_MASK |
						  GDK_POINTER_MOTION_HINT_MASK);

	/* Make the hbox for the buttons */
	hpackbox = gtk_hbox_new(TRUE, 0);

	/* Put the graph in the vpackbox */
	gtk_box_pack_start(GTK_BOX(packbox), graph, TRUE, TRUE, 0);

	/* Put the hpackbox in the vpackbox */
	gtk_box_pack_start(GTK_BOX(packbox), hpackbox, FALSE, FALSE, 5);

	/* make a button */
	clearbutton = gtk_button_new_with_label("Clear");

	/* Link it to the clear function */
	gtk_signal_connect(GTK_OBJECT(clearbutton), "clicked",
					   GTK_SIGNAL_FUNC(clearcallback), NULL);

	/* make the toggles */
	gtbutton = gtk_toggle_button_new_with_label("Graph");
	stbutton = gtk_toggle_button_new_with_label("Spectromap");
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(gtbutton), 1);
	gtk_signal_connect(GTK_OBJECT(gtbutton), "clicked",
					   GTK_SIGNAL_FUNC(graph_toggle_callback), stbutton);
	gtk_signal_connect(GTK_OBJECT(stbutton), "clicked",
					   GTK_SIGNAL_FUNC(graph_toggle_callback), gtbutton);

	/* pack the buttons into the hbox */
	gtk_box_pack_start(GTK_BOX(hpackbox), gtbutton, FALSE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(hpackbox), stbutton, FALSE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(hpackbox), clearbutton, FALSE, TRUE, 0);


	/* Show it */
	gtk_widget_show(graph);

	gtk_widget_show(clearbutton);
	gtk_widget_show(gtbutton);
	gtk_widget_show(stbutton);

	gtk_widget_show(hpackbox);
	gtk_widget_show(packbox);

	gtk_widget_show(window);

	/* Add our input monitoring function */
	gdk_input_add(sockpair[0], GDK_INPUT_READ, ipc_poll_callback, NULL);

	gtk_main();

	kill(ipcpid, SIGKILL);

	return(0);
}
