/* Metageek WiSPY LibUSB interface 
 *
 * LibUSB userspace control for the Metageek WiSPY spectrum analyzer
 * http://www.metageek.net/
 *
 * Basic dumper to dump the raw values from the USB device to stdout.  Meant
 * to be fed to another app via popen (such as the python grapher).  Caller
 * is responsible for converting the RSSI values.
 *
 * Version 1.0
 * Mike Kershaw/Dragorn <dragorn@kismetwireless.net>
 *
 * This code is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Extra thanks to Ryan Woodings @ Metageek for interface documentation
 */

#include <stdio.h>
#include <usb.h>

#include "config.h"

#ifdef HAVE_LINUX_DEVDISCONNECT
/* Needed for our own tools */
#include <unistd.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <linux/usbdevice_fs.h>
#include <errno.h>
#endif

/* First-gen firmware */
#define METAGEEK_WISPY1_VID	0x04b4
#define METAGEEK_WISPY1_PID	0x0bad

/* Second-gen firmware */
#define METAGEEK_WISPY2_VID	0x1781
#define METAGEEK_WISPY2_PID	0x083e

#define TIMEOUT 3000
#define WAITINBETWEEN 500000

/* Hid get report from spec */
#define HID_GET_REPORT 0x01
#define HID_RT_FEATURE 0x03

/* Basic device initialize, grabs either version 1 or version 2 */
static struct usb_device *init_wispy_device() {
	struct usb_bus *bus;
	struct usb_device *dev;

	/* Libusb init */
	usb_init();
	usb_find_busses();
	usb_find_devices();

	/* Scan all the devices on all the busses */
	for (bus = usb_busses; bus; bus = bus->next) {
		for (dev = bus->devices; dev; dev = dev->next) {
			if (((dev->descriptor.idVendor == METAGEEK_WISPY1_VID) &&
				 (dev->descriptor.idProduct == METAGEEK_WISPY1_PID)) ||
				((dev->descriptor.idVendor == METAGEEK_WISPY2_VID) &&
				 (dev->descriptor.idProduct == METAGEEK_WISPY2_PID)))
				return dev;
		}
	}

	return NULL;
}

#ifdef HAVE_LINUX_DEVDISCONNECT
/* Libusb doesn't seem to always provide this, so we'll use our own, taken from the 
* usb_detatch_kernel_driver_np...
*
* THIS IS A HORRIBLE EVIL HACK THAT SHOULDN'T BE DONE, EVER
* 
*/
struct local_usb_ioctl {
	int ifno;
	int ioctl_code;
	void *data;
};

struct ghetto_libusb_devhandle {
	int fd;
	/* Nooo... so bad. */
};

int usb_detach_kernel_driver(struct usb_dev_handle *dev, int interface) {
	struct local_usb_ioctl command;
	struct ghetto_libusb_devhandle *gdev;

	command.ifno = interface;
	command.ioctl_code = USBDEVFS_DISCONNECT;
	command.data = NULL;

	gdev = (struct ghetto_libusb_devhandle *) dev;

	if (ioctl(gdev->fd, USBDEVFS_IOCTL, &command) < 0) {
		fprintf(stderr, "Could not detach kernel driver from interface %d: %s",
			interface, strerror(errno));
		return -1;
	}

	return 0;
}
#endif

int main(int argc, char *argv[]) {
	struct usb_device *dev;
	struct usb_dev_handle *wispy;
	int arg;
	char buf[8];
	int x;
	struct timeval tm;

	memset(buf, 0, sizeof(char) * 8);

	dev = init_wispy_device();
	if (!dev) {
		printf("Unable to find device.\n");
		return 1;
	}

	/* Open the wispy device */
	wispy = usb_open(dev);
	if (!wispy) {
		printf("Unable to open device\n");
		return 1;
	}

#if 1
#ifndef SYS_DARWIN
	/* Grab it like a keyboard interface, interface 0 */
	if (usb_claim_interface(wispy, 0) < 0) {
		printf("Unable to claim interface 0 -- are you root?\n"
			   "Is something else using it (another program or the kernel HID)?\n");
#ifdef HAVE_LINUX_DEVDISCONNECT
		printf("Using terrible hack to try to detach from the Linux HID...\n"
			   "This might not work at all\n");

		if (usb_detach_kernel_driver(wispy, 0) < 0) {
			printf("Failed to detatch from HID.\n");
			exit(1);
		}

		if (usb_claim_interface(wispy, 0) < 0) {
			printf("Seemed to detach from kernel but still couldn't claim interface.\n");
			exit(1);
		}

		printf("Detatched the USB interface from the kernel driver.\n");
#else
		exit(1);
#endif

	}
#endif
#endif

	while (1) {
		/* Grab a report */
		if (usb_control_msg(wispy, USB_ENDPOINT_IN + USB_TYPE_CLASS + USB_RECIP_INTERFACE,
							HID_GET_REPORT, (HID_RT_FEATURE << 8),
							0, buf, 8, TIMEOUT) == 0) {
			printf("Failure in control_msg\n");
			return 1;
		}

		for (x = 0; x < 8; x++) 
			printf("%d ", buf[x]);
		printf("\n");

		tm.tv_sec = 0;
		tm.tv_usec = 8000;

		select(0, NULL, NULL, NULL, &tm);
	}

	usb_close(wispy);

	return 0;
}	

